#include <linux/module.h>
#include <linux/dmaengine.h>
#include <linux/dma-mapping.h>
#include <linux/edma.h>
#include <linux/slab.h>
#include <linux/completion.h>
#include <linux/interrupt.h>

#define TRANSFER_SIZE 64 // 4KB transfer size

static dma_addr_t src_dma, dst_dma;
static char *src_buffer, *dst_buffer;
static struct dma_chan *dma_chan;
static struct completion dma_complete;

// Callback function for the EDMA transfer
static void edma_callback(void *completion)
{
    complete(completion);
}

// Function to set up and initiate EDMA transfer
static int edma_memcpy_transfer(void)
{
    struct dma_async_tx_descriptor *tx;
    dma_cap_mask_t mask;
    int ret;

    // Allocate memory for source and destination buffers
    src_buffer = kmalloc(TRANSFER_SIZE, GFP_KERNEL);
    if (!src_buffer) {
        pr_err("Failed to allocate source buffer\n");
        return -ENOMEM;
    }

    dst_buffer = kmalloc(TRANSFER_SIZE, GFP_KERNEL);
    if (!dst_buffer) {
        pr_err("Failed to allocate destination buffer\n");
        kfree(src_buffer);
        return -ENOMEM;
    }

    // Initialize source buffer with test data
    memset(src_buffer, 0xAA, TRANSFER_SIZE);
    memset(dst_buffer, 0x00, TRANSFER_SIZE);

    // Map memory for DMA (EDMA needs physical addresses)
    src_dma = dma_map_single(NULL, src_buffer, TRANSFER_SIZE, DMA_TO_DEVICE);
    dst_dma = dma_map_single(NULL, dst_buffer, TRANSFER_SIZE, DMA_FROM_DEVICE);

    if (dma_mapping_error(NULL, src_dma) || dma_mapping_error(NULL, dst_dma)) {
        pr_err("Failed to map DMA buffers\n");
        ret = -ENOMEM;
        goto free_buffers;
    }

    // Initialize completion structure
    init_completion(&dma_complete);

    // Request EDMA channel
    dma_cap_zero(mask);
    dma_cap_set(DMA_MEMCPY, mask);
//dma_chan = dma_request_chan(NULL, "edma_memcpy0");
    dma_chan = dma_request_chan(NULL, "dma1chan20");
    if (IS_ERR(dma_chan)) {
        pr_err("Failed to request EDMA channel\n");
        ret = PTR_ERR(dma_chan);
        goto unmap_dma;
    }

    // Prepare the EDMA transfer
    tx = dmaengine_prep_dma_memcpy(dma_chan, dst_dma, src_dma, TRANSFER_SIZE, DMA_CTRL_ACK | DMA_PREP_INTERRUPT);
    if (!tx) {
        pr_err("Failed to prepare EDMA transfer\n");
        ret = -EIO;
        goto release_channel;
    }

    // Set the callback function
    tx->callback = edma_callback;
    tx->callback_param = &dma_complete;

    // Submit the EDMA transfer
    ret = dma_submit_error(dmaengine_submit(tx));
    if (ret) {
        pr_err("Failed to submit EDMA transfer\n");
        goto release_channel;
    }

    // Start the EDMA transfer
    dma_async_issue_pending(dma_chan);

    // Wait for the transfer to complete
    wait_for_completion(&dma_complete);

    pr_info("EDMA transfer completed. Verifying data...\n");

    // Verify if the destination buffer now contains the source buffer data
    if (memcmp(src_buffer, dst_buffer, TRANSFER_SIZE) == 0) {
        pr_info("EDMA transfer verified successfully!\n");
    } else {
        pr_err("EDMA transfer verification failed\n");
    }

release_channel:
    dma_release_channel(dma_chan);

unmap_dma:
    dma_unmap_single(NULL, src_dma, TRANSFER_SIZE, DMA_TO_DEVICE);
    dma_unmap_single(NULL, dst_dma, TRANSFER_SIZE, DMA_FROM_DEVICE);

free_buffers:
    kfree(src_buffer);
    kfree(dst_buffer);

    return ret;
}

// Module initialization
static int __init edma_test_init(void)
{
    pr_info("Initializing EDMA test module\n");
    return edma_memcpy_transfer();
}

// Module exit
static void __exit edma_test_exit(void)
{
    pr_info("Exiting EDMA test module\n");
}

module_init(edma_test_init);
module_exit(edma_test_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Your Name");
MODULE_DESCRIPTION("EDMA memcpy test for AM5748");

